import .Roots ;

//! Finds the minimum of a function using the first derivative of a function
float SimpleQuestWith1stDerivatives(  function df , array range , int NITER , float tol_rel )
{
  return RegulaFalsi( df , range, tol_rel , NITER );
}


//! A simplex is a set of N + 1 points in a N dimensional space. It's one of the most beautiful methods to minimize a function of more than 1 variable.
class Simplex {
  private
  array punts ;
  array cdg  ; //El centre de gravetat del símplex.
  function f ;
  array sumar ( array a1 , array a2 )
  {
    int p = sizeof(a1) , q = sizeof(a2);
    if (p != q )
      {
	throw( ({"The points at the simplex don't have the correct dimension\n" , backtrace() }));
      }
    array resultat = ({});
    for ( int i = 0 ; i < p ; ++i )
      {
	resultat += ({ a1[i] + a2[i] }) ;
      }
    return resultat ;
  }
  
  array restar ( array a1 , array a2 )
  {
    int p = sizeof(a1) , q = sizeof(a2);
    if (p != q )
      {
	throw( ({"The points at the simplex don't have the correct dimension\n" , backtrace() }));
      }
    array resultat = ({});
    for ( int i = 0 ; i < p ; ++i )
      {
	resultat += ({ a1[i] - a2[i] }) ;
      }
    return resultat ;
  }

  array multiplicar ( array a1 , float c )
  {
    int p = sizeof(a1);
    array resultat = ({});
    for ( int i = 0 ; i < p ; ++i )
      {
	resultat += ({ a1[i] * c }) ;
      }
    return resultat ;
  }
  private void sort()
  {
    float aux ;
    array element1 , element2 ;
    int N = sizeof(punts);
    for( int i = 0 ; i < N ; ++i )
      {
	for ( int j = 1 ; j < N ; ++j )
	  {
	    element1 = punts[j-1] ;
	    element2 = punts[j] ;
	    if ( f(@element1) > f(@element2) )
	      {
		punts[j] = element1;
		punts[j-1] = element2 ;
	      }
	  }
      }
    
  }
  
  private array gravityCenter()
  {
    array aux = punts[0];
    int N = sizeof(punts);
    for(int i =  1 ; i < N - 1 ; ++i )
      {
	aux = sumar( aux , punts[i] );
      }
    return multiplicar (aux ,   1.0 / (N - 1 ) ) ;
  }
  
  
  //! To create a Simplex, we give a function to minimize and an array of points. Exemple: if we want to minimize a function of two variables, f(x,y), we must call Simplex with f(x,y) and an array of 3 elements: ({ ({x1, y1 }) , ({x2, y2}) , ({x3, y3}) })
  public void create(function function2minimize , array points )
  {
    punts = points ;
    f = function2minimize ;
  }
  
  
  //! To minimize a Simplex, we can do NITER iterations and tolerate a relative error of tol_rel.
  array minimize( int NITER , float tol_rel )
  {
    int iteracio = 0 , pos_pitjor = sizeof(punts) - 1, pos_millor = 0 , pos_quasipitjor = pos_pitjor - 1;
    float tol ;
    array millor, pitjor, quasipitjor, nou, nou2 , cdg ;
    int N = sizeof(punts);
    while(iteracio < NITER )
      {
	++iteracio ;
	sort();
	millor = punts[pos_millor];
	pitjor = punts[pos_pitjor];
	quasipitjor = punts[pos_quasipitjor];
	cdg = gravityCenter();
	nou = restar( multiplicar( cdg , 2.0 ) , pitjor ) ;
	if ( f(@nou) < f(@millor) )
	  {
	    nou2 = restar (multiplicar( cdg , 3.0 ) , pitjor ) ;
	    //Cas 1a : el nou punt és millor que nou -> canviem pitjor per nou2 
	    if ( f(@nou2 ) < f(@nou) )
	      {
		punts[pos_pitjor] = nou2 ;
	      }
	    //Cas 1b: el nou punt és pitjor que nou -> canviem pitjor per nou
	    else
	      {
		punts[pos_pitjor] = nou ;
	      }
	  }
	//Cas 2: el nou punt és millor que quasipitjor -> canviem nou per quasipitjor
	if ( f(@nou) < f(@quasipitjor) )
	  {
	    punts[pos_pitjor] = nou ;
	  }
	//Cas 3: nou és pitjor que quasipitjor
	else
	  {
	    nou2 =  multiplicar(sumar(cdg , pitjor ), 0.5 );
	    //Cas 3a: el nou punt és millor que quasipitjor -> canviem pitjor per nou2
	    if ( f(@nou2 ) < f(@quasipitjor) )
	      {
		punts[pos_pitjor] = nou2;
	      }
	    //Cas 3b: el nou punt és pitjor que quasipitjor -> contracció global del símplex
	    else
	      {
		for( int auxiliar = 1 ; auxiliar < N ; ++auxiliar )
		  {
		    punts[auxiliar] = multiplicar(restar(millor, punts[auxiliar]), 0.75 ) ;
		  }
	      }
	      
	  }
	sort();
	tol = max ( (f(@pitjor) + f(@millor)) * tol_rel / 2.0 , tol_rel ) ;
	if ( tol >= abs( f(@millor) - f(@pitjor) ) )
	  {
	    return millor ;
	  }
      }
    throw( ({"Not minimum found\n" , backtrace() }) );
  }
}

int main ()
{
  function f = lambda ( float x , float y ) { return pow(x  - 1.0 , 2.0) + pow(y - 1.0  , 2.0 ) ; } ;
  array points = ({ ({ 2.0, 1.0}) , ({ 2.0 , 5.0 }) , ({ 2.0, -1.0 }) }) ;
  Simplex smp = Simplex( f , points );
  //We try to minimize the sin :
  float resultat = SimpleQuestWith1stDerivatives( cos , ({ 0.0, 3.75 }) , 100 , 0.01 ) ;
  write("The function sin(x) has an extremal in x = " + resultat + "\n");
  write("Now minimizing simplex...\n");
  array resultat2 = smp->minimize( 1000 , 0.0001 );
  write("The minimum of the function  (x - 1)^2 + (y - 1)^2 should be in (1.0 , 1.0) and we find It in x = " + resultat2[0] + " , y = " + resultat2[1] + "\n" );
}



